from datetime import datetime

import numpy as np
import pandas as pd
import plotly.offline as py
from plotly.graph_objs import *

py.init_notebook_mode()


def consolidate_data(detections):
    """

    Takes set of detections, cleans and sumarises the detections for the
    timeline.

    :param detections: A Pandas DataFrame of animal detections

    :return: A Pandas DataFrame of catalognumber, station, date, latitude,
        longitude and detection counts by day
    """

    detections = detections.copy(deep=True)
    if 'receiver' in detections.columns:
        detections = detections[~(detections.receiver == 'release')]
    detections['date'] = pd.to_datetime(detections.datecollected).dt.date
    detections['det_counts'] = 0
    detections = detections.groupby(
        ['catalognumber', 'date', 'station'], as_index=False).agg({
            'det_counts': 'count',
            'latitude': 'mean',
            'longitude': 'mean',
        })[['catalognumber',
            'date',
            'station',
            'det_counts',
            'latitude',
            'longitude']]
    detections.det_counts = (detections.det_counts / 10.0) + 5
    detections['date'] = pd.to_datetime(detections['date'])

    detections.sort_values('date', inplace=True)

    catalognumbers = pd.DataFrame(
        detections.catalognumber.unique(), columns=['catalognumber'])
    catalognumbers['color'] = catalognumbers.index + 1

    detections = detections.merge(catalognumbers, on='catalognumber')
    return detections


def create_grid(detections):
    """

    Takes the a set of consolidaed detections (the output from
    ``consolidate_data()``) and organizes them into a Plotly grid like format.

    :param detections: A Pandas DataFrame of catalognumber, station, date,
        latitude, longitude and detection counts by day

    :return: A Plotly grid like dataframe
    """

    total_grid = pd.DataFrame()
    for date, data in detections.groupby('date'):
        grid = pd.DataFrame()
        grid['x-' + str(date.date())] = data.longitude
        grid['y-' + str(date.date())] = data.latitude
        grid['size-' + str(date.date())] = data.det_counts
        grid['station-' + str(date.date())] = data.station
        grid['color-' + str(date.date())] = data.color
        grid['catalognumber-' + str(date.date())] = data.catalognumber
        grid.reset_index(inplace=True, drop=True)
        total_grid = pd.concat([total_grid, grid], ignore_index=False, axis=1)
    return total_grid


def create_trace(detections, total_grid, is_mapbox=False,
                 colorscale='Rainbow'):
    """

    :param detections: A Pandas DataFrame of catalognumber, station, date,
        latitude, longitude and detection counts by day
    :param total_grid: A Pandas DataFrame from ``create_grid()``
    :param is_mapbox: A boolean indicating whether to return a Scattermapbox or
        Scattergeo trace
    :param colorscale: A string to indicate the color index. See here for
        options:
        https://community.plot.ly/t/what-colorscales-are-available-in-plotly-and-which-are-the-default/2079

    :return: a Plotly Scattergeo or Scattermapbox trace of the first frame
    """
    trace = dict(
        lon=total_grid['x-' + str(detections.date.min().date())].dropna(),
        lat=total_grid['y-' + str(detections.date.min().date())].dropna(),
        text=total_grid['catalognumber-' +
                        str(detections.date.min().date())].dropna(),
        mode="markers",
        hoverinfo="lon+lat+text",
        marker=dict(
            size=total_grid['size-' +
                            str(detections.date.min().date())].dropna(),
            color=total_grid['color-' +
                             str(detections.date.min().date())].dropna(),
            cmin=detections.color.min(),
            cmax=detections.color.max(),
            autocolorscale=False,
            colorscale=colorscale
        )

    )
    if is_mapbox:
        trace = Scattermapbox(trace)
    else:
        trace = Scattergeo(trace)
    return trace


def create_frames(detections, total_grid, is_mapbox=False):
    """

    :param detections: A Pandas DataFrame of catalognumber, station, date,
        latitude, longitude and detection counts by day
    :param total_grid: A Pandas DataFrame from ``create_grid()``
    :param is_mapbox: A boolean indicating whether to return a Scattermapbox or
        Scattergeo trace

    :return: An array of Plotly Frames
    """
    frames = []
    for date in pd.date_range(detections.date.min(), detections.date.max()):
        date = date.date()

        if 'x-' + str(date) in total_grid.columns:
            frame_trace = dict(
                lon=total_grid['x-' + str(date)].dropna(),
                lat=total_grid['y-' + str(date)].dropna(),
                text=total_grid['catalognumber-' + str(date)].dropna(),
                mode="markers",
                hoverinfo="lon+lat+text",
                marker=dict(
                    size=total_grid['size-' + str(date)].dropna(),
                    color=total_grid['color-' + str(date)].dropna()
                )
            )
        else:
            frame_trace = dict(
                lon=[0],
                lat=[0],
                text=[0],
                mode="markers",
                marker=dict(
                    size=[0]
                )
            )

        if is_mapbox:
            date_trace = Scattermapbox(frame_trace)
        else:
            date_trace = Scattergeo(frame_trace)
        frame = Frame(
            name=str(date),
            data=[date_trace],
            traces=[0]
        )

        frames.append(frame)
    return frames


def define_updatemenus(animation_time=1000, transition_time=300):
    """
    :param animation_time: The amount of time in milliseconds for each frame
    :param transition_time: The smount of time in milliseconds between frames

    :return: A dictionary for updatemenus and buttons in Plotly
    """
    updatemenus = dict(
        # GENERAL
        type="buttons",
        showactive=False,
        x=0.1,
        y=0,
        pad=dict(t=60, r=10),
        xanchor="right",
        yanchor="top",
        direction="left",

        # Buttons
        buttons=[
            dict(
                method="animate",
                label="Play",

                args=[
                    None,
                    dict(
                        # False quicker but disables animations
                        frame=dict(duration=animation_time, redraw=False),
                        fromcurrent=True,
                        # easing = "cubic-in-out"
                        transition=dict(duration=transition_time,
                                        easing="quadratic-in-out"),
                        mode="immediate",
                    ),
                ],
            ),
            dict(
                method="animate",
                label="Pause",

                # PAUSE
                args=[
                    [None],  # Note the list
                    dict(
                        frame=dict(duration=0, redraw=False),  # Idem
                        mode="immediate",
                        transition=dict(duration=0),
                    ),
                ],
            ),
        ],
    )
    return updatemenus


def define_sliders(detections, animation_time=300, slider_transition_time=300):
    """

    :param detections: A Pandas DataFrame of catalognumber, station, date,
        latitude, longitude and detection counts by day
    :param transition_time: The amount of time in milliseconds between frames
    :param slider_transition_time: The amount of time in milliseconds between
        frames for the slider

    :return: A Plotly sliders dictionary
    """
    sliders = dict(
        active=0,
        steps=[],

        currentvalue=dict(
            font=dict(size=16),
            prefix="Year : ",
            xanchor="right",
            visible=True,
        ),
        transition=dict(
            duration=slider_transition_time,
            easing="cubic-in-out",
        ),

        # PLACEMENT
        x=0.1,
        y=0,
        pad=dict(t=40, b=10),
        len=0.9,
        xanchor="left",
        yanchor="top",
    )

    for date in pd.date_range(detections.date.min(), detections.date.max()):
        date = date.date()
        slider_step = dict(

            # GENERAL
            method="animate",
            value=str(date),
            label=str(date),

            # ARGUMENTS
            args=[
                [str(date)],
                dict(
                    frame=dict(duration=animation_time, redraw=False),
                    transition=dict(duration=slider_transition_time),
                    mode="immediate",
                ),
            ],

        )

        sliders["steps"].append(slider_step)
    return sliders


def define_layout(detections, title, plotly_geo=None,  mapbox_token=None,
                  style='light'):
    """

    :param detections: A Pandas DataFrame of catalognumber, station, date,
        latitude, longitude and detection counts by day
    :param title: the title of the plot
    :param plotly_geo: an optional dictionary to control the
        geographic aspects of the plot
    :param mapbox_token: A string of mapbox access token
    :param style: The style for the Mapbox tileset:
        https://plot.ly/python/reference/#layout-mapbox-style

    :return: A plotly layout dictionary
    """

    if mapbox_token is None:
        if plotly_geo is None:
            plotly_geo = dict(
                showland=True,
                landcolor="rgb(255, 255, 255)",
                showocean=True,
                oceancolor="rgb(212,212,212)",
                showlakes=True,
                lakecolor="rgb(212,212,212)",
                showrivers=True,
                rivercolor="rgb(212,212,212)",
                resolution=50,
                showcoastlines=False,
                showframe=False,
                projection=dict(
                    type='mercator',
                )
            )
        plotly_geo.update(
            center=dict(
                lon=detections.longitude.mean(),
                lat=detections.latitude.mean()
            ),
            lonaxis=dict(
                range=[detections.longitude.min(), detections.longitude.max()],
            ),
            lataxis=dict(
                range=[detections.latitude.min(), detections.latitude.max()],
            )
        )
        layout = dict(
            geo=plotly_geo,
            title=title,
            hovermode='closest',
            showlegend=False,
        )
    else:
        mapbox = dict(
            accesstoken=mapbox_token,
            center=dict(
                lon=detections.longitude.mean(),
                lat=detections.latitude.mean()
            ),
            zoom=5,
            style=style
        )
        layout = dict(title=title,
                      autosize=True,
                      hovermode='closest',
                      showlegend=False,
                      mapbox=mapbox
                      )
    return layout


def timeline(detections, title='Timeline', height=700, width=1000,
             ipython_display=True, mapbox_token=None, plotly_geo=None,
             animation_time=1000, transition_time=300,
             slider_transition_time=300, colorscale='Rainbow', style='light'):
    """

    :param detections: A Pandas DataFrame of catalognumber, station, date,
        latitude, longitude and detection counts by day
    :param ipython_display: a boolean to show in a notebook
    :param title: the title of the plot
    :param height: the height of the plotly
    :param width: the width of the plotly
    :param plotly_geo: an optional dictionary to control the
        geographic aspects of the plot
    :param mapbox_token: A string of mapbox access token
    :param animation_time: The amount of time in milliseconds for each frame
    :param transition_time: The amount of time in milliseconds between frames
    :param slider_transition_time: The amount of time in milliseconds between
        frames for the slider
    :param colorscale: A string to indicate the color index. See here for
        options:
        https://community.plot.ly/t/what-colorscales-are-available-in-plotly-and-which-are-the-default/2079
    :param style: The style for the Mapbox tileset:
        https://plot.ly/python/reference/#layout-mapbox-style

    """
    detections = consolidate_data(detections)
    total_grid = create_grid(detections)
    if mapbox_token is None:
        trace = create_trace(detections, total_grid, colorscale=colorscale)
        frames = create_frames(detections, total_grid)
    else:
        trace = create_trace(detections, total_grid,
                             is_mapbox=True, colorscale=colorscale)
        frames = create_frames(detections, total_grid, is_mapbox=True)
    updatemenus = define_updatemenus(animation_time, transition_time)
    sliders = define_sliders(detections, animation_time,
                             slider_transition_time)
    layout = define_layout(detections, title, mapbox_token=mapbox_token,
                           style=style)

    layout.update(dict(
        updatemenus=[updatemenus],
        sliders=[sliders],
    ))

    layout.update(
        height=height,
        width=width
    )

    if ipython_display:
        fig = {'data': [trace], 'layout': layout, 'frames': frames}

        py.init_notebook_mode()
        return py.iplot(fig)
    else:
        fig = {'data': [trace], 'layout': layout, 'frames': frames}
        return py.plot(fig, filename="{}.html".format(
            title.lower().replace(" ", "_")))



from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from geopy.distance import geodesic

from resonate.library.exceptions import GenericException


def get_distance_matrix(detections):
    """
    Creates a distance matrix of all stations in the array or line.

    :param detections: a Pandas DataFrame of detections

    :return: A Pandas DataFrame matrix of station to station distances

    """
    stn_grouped = detections.groupby('station')
    stn_locs = stn_grouped[['longitude', 'latitude']].mean()

    dist_mtx = pd.DataFrame(
        np.zeros(len(stn_locs) ** 2).reshape(len(stn_locs), len(stn_locs)),
        index=stn_locs.index, columns=stn_locs.index)

    for cstation in dist_mtx.columns:
        for rstation in dist_mtx.index:
            cpoint = (stn_locs.loc[cstation, 'latitude'],
                      stn_locs.loc[cstation, 'longitude'])
            rpoint = (stn_locs.loc[rstation, 'latitude'],
                      stn_locs.loc[rstation, 'longitude'])
            dist_mtx.loc[rstation, cstation] = geodesic(cpoint, rpoint).m
    dist_mtx.index.name = None
    return dist_mtx


def filter_detections(detections, suspect_file=None,
                      min_time_buffer=3600,
                      distance_matrix=False):
    """
    Filters isolated detections that are more than min_time_buffer apart from
    other dets. for a series of detections in detection_file. Returns Filtered
    and Suspect dataframes.
    suspect_file can be a file of existing suspect detections to remove before
    filtering.
    dist_matrix is created as a matrix of between-station distances from
    stations defined in the input file.

    :param detections: A Pandas DataFrame of acoustic detections

    :param suspect_file: Path to a user specified suspect file, same format as the detections

    :param min_time_buffer: The minimum of time required for outlier detections
        in seconds

    :param distance_matrix: A boolean of whether or not to generate the
        distance matrix

    :return: A list of Pandas DataFrames of filtered detections, suspect
        detections, and a distance matrix

    """

    # Set of mandatory column names for detection_file
    mandatory_columns = set(['station',
                             'unqdetecid',
                             'datecollected',
                             'catalognumber'])

    # Subtract all detections found in the user defined suspect file
    if suspect_file:
        print("Found suspect file {0}. Subtracting detections from input".format(
            suspect_file))
        susp_dets = pd.read_csv(suspect_file)
        good_dets = pd.concat([detections, susp_dets], ignore_index=True)
        good_dets.drop_duplicates(mandatory_columns, keep=False, inplace=True)

    elif mandatory_columns.issubset(detections.columns):
        # calculate detections to filtered

        # For each individual catalognumber:
        # Determine the space between each detection.
        # If the space before + after > min_time_buffer
        # Remove that detection row from the detections and add it to suspect detections.
        # SQL that does this is in load_to_postgresql under createSuspect
        detections = detections.copy(deep=True)
        ind = detections['catalognumber'].unique()
        detections.loc[:, 'datecollected'] = pd.to_datetime(
            detections['datecollected'])
        user_int = timedelta(seconds=min_time_buffer)
        good_dets = pd.DataFrame()
        susp_dets = pd.DataFrame()
        grouped = detections.groupby('catalognumber')
        for anm in ind:
            anm_dets = grouped.get_group(anm).sort_values(
                'datecollected', ascending=True)
            intervals = anm_dets['datecollected'] - \
                anm_dets['datecollected'].shift(1)
            post_intervals = anm_dets['datecollected'].shift(
                -1) - anm_dets['datecollected']

            good_dets = good_dets.append(
                anm_dets[
                    (intervals <= user_int) | (post_intervals <= user_int)
                ]
            )

        # If they aren't a good det, they're suspect!
        # TODO: Reporting: Decide if we want to report the big 'before/after'
        # triplicate in Suspect Dets
        # If so, building susp_dets gets tougher, involves a merge and then a
        # append.
        # For now, just a matter of putting the complement of the good dets in
        # the susp_dets
        susp_dets = detections[~detections['unqdetecid'].isin(
            good_dets['unqdetecid'])]

    else:
        raise GenericException("Missing required input columns: {}".format(
            mandatory_columns - set(detections.columns)))

    print("Total detections in filtered dataframe: {0}".format(
        len(good_dets.index)))
    print("{0} suspect detections removed".format(len(susp_dets.index)))

    output_dict = {"filtered": good_dets, "suspect": susp_dets}

    if distance_matrix:

        # Must now have lat and long columns as well.
        dm_mandatory_columns = set(['latitude', 'longitude'])
        if dm_mandatory_columns.issubset(detections.columns):
            output_dict['dist_mtrx'] = get_distance_matrix(detections)
            print("There are {0} station locations in the distance \
                matrix".format(len(output_dict['dist_mtrx'].index)))
        else:
            raise GenericException("Missing required input columns for \
                distance matrix calc: {}".format(
                dm_mandatory_columns - set(detections.columns)))

    return output_dict


def distance_filter(detections, maximum_distance=100000):
    """
    :param detections: a Pandas DataFrame of acoustic detection
    :param maximum_distance: a umber in meters, default is 100000

    :return: A list of Pandas DataFrames of filtered detections and suspect
        detections
    """
    pd.options.mode.chained_assignment = None

    mandatory_columns = set(['station',
                             'unqdetecid',
                             'datecollected',
                             'catalognumber'])

    if mandatory_columns.issubset(detections.columns):
        dm = get_distance_matrix(detections)

        lead_lag_stn_df = pd.DataFrame()

        for index, group in detections.sort_values(['datecollected']).groupby(['catalognumber']):
            group['lag_station'] = group.station.shift(1).fillna(group.station)
            group['lead_station'] = group.station.shift(
                -1).fillna(group.station)
            lead_lag_stn_df = lead_lag_stn_df.append(group)

        del detections

        distance_df = pd.DataFrame()

        for index, group in lead_lag_stn_df.groupby(['station', 'lag_station', 'lead_station']):
            stn = group.station.unique()[0]
            lag_stn = group.lag_station.unique()[0]
            lead_stn = group.lead_station.unique()[0]
            lag_distance = dm.loc[stn, lag_stn]
            lead_distance = dm.loc[stn, lead_stn]
            group['lag_distance_m'] = lag_distance
            group['lead_distance_m'] = lead_distance
            distance_df = distance_df.append(group)

        del lead_lag_stn_df
        distance_df.sort_index(inplace=True)

        filtered_detections = dict()
        filtered_detections['filtered'] = distance_df[(distance_df.lag_distance_m <= maximum_distance) & (
            distance_df.lead_distance_m <= maximum_distance)].reset_index(drop=True)
        filtered_detections['suspect'] = distance_df[(distance_df.lag_distance_m > maximum_distance) | (
            distance_df.lead_distance_m > maximum_distance)].reset_index(drop=True)
        return filtered_detections
    else:
        raise GenericException("Missing required input columns: {}".format(
            mandatory_columns - set(detections.columns)))


def velocity_filter(detections, maximum_velocity=10):
    """
    :param detections:
    :param maximum_velocity:

    :return: A list of Pandas DataFrames of filtered detections and suspect
        detections
    """
    pd.options.mode.chained_assignment = None

    mandatory_columns = set(['station',
                             'unqdetecid',
                             'datecollected',
                             'catalognumber'])

    if mandatory_columns.issubset(detections.columns):

        dm = get_distance_matrix(detections)

        lead_lag_df = pd.DataFrame()

        for index, group in detections.sort_values(['datecollected']).groupby(['catalognumber']):
            group['lag_station'] = group.station.shift(1).fillna(group.station)
            group['lead_station'] = group.station.shift(
                -1).fillna(group.station)

            group.datecollected = pd.to_datetime(group.datecollected)
            group['lag_time_diff'] = group.datecollected.diff().fillna(
                timedelta(seconds=1))
            group['lead_time_diff'] = group.lag_time_diff.shift(
                -1).fillna(timedelta(seconds=1))

            lead_lag_df = lead_lag_df.append(group)

        del detections

        vel_df = pd.DataFrame()
        for index, group in lead_lag_df.groupby(['station', 'lag_station', 'lead_station']):
            stn = group.station.unique()[0]
            lag_stn = group.lag_station.unique()[0]
            lead_stn = group.lead_station.unique()[0]

            lag_distance = dm.loc[stn, lag_stn]
            lead_distance = dm.loc[stn, lead_stn]

            group['lag_distance_m'] = lag_distance
            group['lead_distance_m'] = lead_distance
            vel_df = vel_df.append(group)

        del lead_lag_df

        vel_df['lag_velocity'] = vel_df.lag_distance_m / \
            vel_df.lag_time_diff.dt.total_seconds()
        vel_df['lead_velocity'] = vel_df.lead_distance_m / \
            vel_df.lead_time_diff.dt.total_seconds()
        vel_df.sort_index(inplace=True)
        filtered_detections = dict()
        filtered_detections['filtered'] = vel_df[(vel_df.lag_velocity <= maximum_velocity) & (
            vel_df.lead_velocity <= maximum_velocity)].reset_index(drop=True)
        filtered_detections['suspect'] = vel_df[(vel_df.lag_velocity > maximum_velocity) | (
            vel_df.lead_velocity > maximum_velocity)].reset_index(drop=True)
        return filtered_detections
    else:
        raise GenericException("Missing required input columns: {}".format(
            mandatory_columns - set(detections.columns)))



import datetime

import numpy as np
import pandas as pd
from resonate.library.exceptions import GenericException


def REI(detections, deployments):
    '''
    Calculates a returns a list of each station and the REI (defined here):

    :param detections: a pandas DataFrame of detections
    :param deployments: a pandas DataFrame of station deployment histories

    :return: a pandas DataFrame of station, REI, latitude, and longitude
    '''

    # Check for proper dataframe and the correct columns
    if not isinstance(detections, pd.DataFrame):
        raise GenericException('input parameter must be a Pandas dataframe')

    mandatory_detection_columns = set(
        ['datecollected', 'fieldnumber', 'scientificname', 'station'])
    mandatory_deployment_columns = set(
        ['station_name', 'deploy_date', 'recovery_date', 'last_download'])

    if mandatory_detection_columns.issubset(detections.columns) and \
            mandatory_deployment_columns.issubset(deployments.columns):

        # Copy and change the deployments to create dates in the 3 mandatory
        # date columns
        deployments = deployments.copy(deep=True)
        detections = detections.copy(deep=True)
        if deployments.recovery_date.dtype != np.dtype('<M8[ns]'):
            deployments['recovery_notes'] = deployments.recovery_date.str.extract(
                r'([A-Za-z\//:]+)', expand=False)
            deployments.recovery_date = deployments.recovery_date.str.extract(
                r'(\d+-\d+-\d+)', expand=False)
            deployments = deployments.replace('-', np.nan)
        deployments.loc[deployments.recovery_date.isnull(
        ), 'recovery_date'] = deployments.last_download

        deployments = deployments[~deployments.recovery_date.isnull()]
        # Cast the date columns to a datetime
        deployments.deploy_date = pd.to_datetime(deployments.deploy_date)
        deployments.recovery_date = pd.to_datetime(deployments.recovery_date)
        deployments.last_download = pd.to_datetime(deployments.last_download)

        # Calculate each receivers total days deployed
        deployments['days_deployed'] = deployments.recovery_date - \
            deployments.deploy_date

        days_active = deployments.groupby('station_name').agg(
            {'days_deployed': 'sum'}).reset_index()
        days_active.set_index('station_name', inplace=True)
        # Exclude all detections that are not registered with receivers in the
        # deployments
        detections = detections[detections.station.isin(
            deployments.station_name)]

        # Calculate array counts and initialize the dataframe
        array_unique_tags = len(detections.fieldnumber.unique())
        array_unique_species = len(detections.scientificname.unique())
        days_with_detections = len(pd.to_datetime(
            detections.datecollected).dt.date.unique())
        array_days_active = (max(deployments.last_download.fillna(deployments.deploy_date.min()).max(
        ), deployments.recovery_date.max()) - min(deployments.deploy_date)).days

        station_reis = pd.DataFrame(columns=['station', 'rei'])

        # Loop through each station in the detections and Calculate REI for
        #  oeach station
        detections.datecollected = pd.to_datetime(
            detections.datecollected).dt.date

        # Loop through each station in the detections and Calculate REI for
        #  oeach station
        for name, data in detections.groupby('station'):
            receiver_unique_tags = len(data.fieldnumber.unique())
            receiver_unique_species = len(data.scientificname.unique())
            receiver_days_with_detections = len(
                pd.to_datetime(data.datecollected).dt.date.unique())

            if name in days_active.index:
                receiver_days_active = days_active.loc[name].days_deployed.days
                if receiver_days_active > 0:
                    rei = (receiver_unique_tags / array_unique_tags) * \
                        (receiver_unique_species / array_unique_species) * \
                        (receiver_days_with_detections / days_with_detections) * \
                        (array_days_active / receiver_days_active)
                    station_reis = station_reis.append({
                        'station': name,
                        'rei': rei,
                        'latitude': data.latitude.mean(),
                        'longitude': data.longitude.mean()},
                        ignore_index=True)
            else:
                print("No valid deployment record for " + name)

        # Normalize REIs to value from 0 to 1
        station_reis.rei = station_reis.rei / station_reis.rei.sum()

        # Cleanup and return the station REI's
        del deployments
        return station_reis
    else:
        raise GenericException("Missing required input columns: {}".format(
            mandatory_detection_columns - set(detections.columns)))
